"""
Abstract websocket functionality.

"""

from __future__ import print_function
from __future__ import unicode_literals

from base64 import b64encode
from hashlib import sha1
import logging
import os

import six
from six.moves.urllib.parse import urlparse

from . import constants
from . import errors
from . import events
from .frame import Frame
from .opcode import Opcode
from .response import Response
from .stream import WebsocketStream
from .session import WebsocketSession
from .status import Status


log = logging.getLogger('lomond')


class WebSocket(object):
    """IO independent websocket functionality."""

    class State(object):
        def __init__(self):
            self.stream = WebsocketStream()
            self.session = None
            self.key = b64encode(os.urandom(16))
            self.sent_request = False
            self.closing = False
            self.closed = False

    def __init__(self, url, protocols=None, agent=None):
        self.url = url
        self.protocols = protocols or []
        self.agent = agent or constants.USER_AGENT

        _url = urlparse(url)
        self.scheme = _url.scheme
        host, _, port = _url.netloc.partition(':')
        if not port:
            port = '443' if self.scheme == 'wss' else '80'
        if not port.isdigit():
            raise ValueError('illegal port value')
        port = int(port)
        self.host = host
        self.port = port
        self._host_port = "{}:{}".format(host, port)
        self.resource = _url.path or '/'
        if _url.query:
            self.resource = "{}?{}".format(self.resource, _url.query)

        self.state = self.State()

    def __repr__(self):
        return "WebSocket('{}')".format(self.url)

    @property
    def is_secure(self):
        return self.scheme == 'wss'

    @property
    def is_closing(self):
        return self.state.closing

    @property
    def is_closed(self):
        return self.state.closed

    @property
    def stream(self):
        return self.state.stream

    @property
    def session(self):
        return self.state.session

    @property
    def key(self):
        return self.state.key

    def connect(self,
                session_class=WebsocketSession,
                poll=5,
                ping_rate=30):
        """Connect the websocket to a session."""
        self.reset()
        self.state.session = WebsocketSession(self)
        return self.session.run(poll=poll, ping_rate=ping_rate)

    def reset(self):
        """Reset the state."""
        self.state = self.State()

    __iter__ = connect

    def close(self, code=None, reason=None):
        """Close the websocket."""
        if self.is_closed:
            log.debug('%r already closed', self)
        else:
            if code is None:
                code = Status.NORMAL
            if reason is None:
                reason = b'goodbye'
            self._send_close(code, reason)
            self.state.closing = True

    def _on_close(self, message):
        """Close logic generator."""
        if message.code in Status.invalid_codes:
            raise errors.ProtocolError(
                'reserved close code ({})',
                message.code
            )
        if self.is_closed:
            return
        if self.is_closing:
            yield events.Closed(message.code, message.reason)
            self.state.closing = False
            self.state.closed = True
        else:
            self.close(message.code, message.reason)
            self.state.closing = True

    def disconnect(self):
        """Disconnect the websocket."""
        self.state.closing = False
        self.state.closed = True

    def feed(self, data):
        """Feed with data from the socket."""
        if self.is_closed:
            return
        try:
            session = self.session
            for message in self.stream.feed(data):
                if isinstance(message, Response):
                    response = message
                    try:
                        protocol, extensions = self.on_response(response)
                    except errors.HandshakeError as error:
                        self.disconnect()
                        yield events.Rejected(response, str(error))
                        break
                    else:
                        yield events.Ready(response, protocol, extensions)
                else:
                    if message.is_close:
                        for event in self._on_close(message):
                            yield event
                    elif message.is_ping:
                        session.send(Opcode.PONG, message.data)
                    elif message.is_pong:
                        yield events.Pong(message.data)
                    elif message.is_binary:
                        yield events.Binary(message.data)
                    elif message.is_text:
                        yield events.Text(message.text)
                    else:
                        yield events.UnknownMessage(message)
                if self.is_closed:
                    break

        except errors.CriticalProtocolError as error:
            # An error that warrants an immediate disconnect.
            # Usually invalid unicode.
            log.debug('critical protocol error; %s', error)
            self.disconnect()

        except errors.ProtocolError as error:
            # A violation of the protocol that allows for a graceful
            # disconnect.
            log.debug('protocol error; %s', error)
            self.close(Status.PROTOCOL_ERROR, six.text_type(error))
            self.disconnect()

        except GeneratorExit:
            log.warn('disconnecting websocket')
            self.disconnect()

    def get_request(self):
        """Get the request (in bytes)"""
        request = [
            "GET {} HTTP/1.1".format(self.resource).encode('utf-8')
        ]
        protocols = ", ".join(self.protocols)
        version = '{}'.format(constants.WS_VERSION)
        headers = [
            (b'Host', self._host_port.encode('utf-8')),
            (b'Upgrade', b'websocket'),
            (b'Connection', b'Upgrade'),
            (b'Sec-WebSocket-Protocol', protocols.encode('utf-8')),
            (b'Sec-WebSocket-Key', self.key),
            (b'Sec-WebSocket-Version', version.encode('utf-8')),
            (b'User-Agent', self.agent.encode('utf-8')),
        ]
        for header, value in headers:
            request.append(header + b': ' + value)
        request.append(b'\r\n')
        request_bytes = b'\r\n'.join(request)
        return request_bytes

    def on_response(self, response):
        """Called when the HTTP response has been received."""

        if response.status_code != 101:
            raise errors.HandshakeError(
                'Websocket upgrade failed (code={})',
                response.status_code
            )

        upgrade_header = response.get(b'upgrade', b'?').lower()
        if upgrade_header != b'websocket':
            raise errors.HandshakeError(
                "Can't upgrade to {}",
                upgrade_header.decode('utf-8', errors='replace')
            )

        accept_header = response.get(b'sec-websocket-accept', None)
        if accept_header is None:
            raise errors.HandshakeError(
                "No Sec-WebSocket-Accept header"
            )

        challenge = b64encode(
            sha1(self.key + constants.WS_KEY).digest()
        )

        if accept_header.lower() != challenge.lower():
            raise errors.HandshakeError(
                "Sec-WebSocket-Accept challenge failed"
            )

        protocol = response.get(b'sec-websocket-protocol')
        extensions = set(response.get_list(b'sec-websocket-extensions'))
        return protocol, extensions

    def send_ping(self, data=b''):
        """Send a ping request."""
        if not isinstance(data, bytes):
            raise TypeError('data argument must be bytes')
        if len(data) > 125:
            raise ValueError('ping data should be <= 125 bytes')
        self.session.send(Opcode.PING, data)

    def send_pong(self, data):
        """Send a ping request."""
        if not isinstance(data, bytes):
            raise TypeError('data argument must be bytes')
        if len(data) > 125:
            raise ValueError('ping data should be <= 125 bytes')
        self.session.send(Opcode.PONG, data)

    def send_binary(self, data):
        """Send a binary frame."""
        if not isinstance(data, bytes):
            raise TypeError('data argument must be bytes')
        self.session.send(Opcode.BINARY, data)

    def send_text(self, text):
        """Send a text frame."""
        if not isinstance(text, six.text_type):
            raise TypeError('text argument must be bytes')
        self.session.send(Opcode.TEXT, text.encode('utf-8'))

    def _send_close(self, code, reason):
        """Send a close frame."""
        frame_bytes = Frame.build_close_payload(code, reason)
        try:
            self.session.send(Opcode.CLOSE, frame_bytes)
        except (errors.WebSocketUnavailable, errors.TransportFail):
            return False
        else:
            return True



"""
Functions related to masking Websocket frames.
https://tools.ietf.org/html/rfc6455#section-5.3

"""

import os
from functools import partial
from itertools import cycle

import six


try:
    from wsaccel.xormask import XorMaskerSimple
except ImportError:
    XorMaskerSimple = None


make_masking_key = partial(os.urandom, 4)

if XorMaskerSimple is not None:
    # Fast C version (works with Py2 and Py3)
    def mask(masking_key, data):
        return XorMaskerSimple(masking_key).process(data)

elif six.PY2:
    # Python 2 compatible version
    from itertools import izip
    def mask(masking_key, data):
        """XOR mask bytes."""
        return b''.join(
            chr(ord(a) ^ ord(b))
            for a, b in izip(cycle(masking_key), data)
        )
else:
    # Can't deny the Py3 version is nicer
    def mask(masking_key, data):
        """XOR mask bytes."""
        return bytes(a ^ b for a, b in zip(cycle(masking_key), data))


if __name__ == "__main__":

    plain = b"Hello, World!"
    key = make_masking_key()

    masked = mask(key, plain)
    print(masked)

    plain = mask(key, masked)
    print(plain)



"""
Coroutine base stream parser.

This little but of 'magic' makes parsing a stream of bytes from a socket
a relatively pain-free exercise.

"""

from __future__ import unicode_literals


class Parser(object):
    """
    Coroutine based stream parser.

    Splits a steam of arbitrary sequences of bytes in to logical
    objects. Essentially the `feed` method will yield any results of
    parsing.

        while True:
            data = sock.recv(1024)
            for obj in parser.feed(data):
                self.on_obj(obj)

    The coroutine magic allows parsers to be non-blocking while still
    implemented in a simple procedural manner.

    """

    def __init__(self):
        self._gen = None
        self._awaiting = None
        self._buffer = []  # Buffer for reads
        self._until = b''  # Buffer for read untils
        self._closed = False
        self.reset()

    class _Awaitable(object):
        """An operation that effectively suspends the coroutine."""
        # Analogous to Python3 asyncio concept

    class _ReadBytes(_Awaitable):
        """Reads a fixed number of bytes."""
        __slots__ = ['remaining']
        def __init__(self, count):
            self.remaining = count

    class _ReadUntil(_Awaitable):
        """Read until a separator."""
        __slots__ = ['sep']
        def __init__(self, sep):
            self.sep = sep

    read = _ReadBytes
    read_until = _ReadUntil

    def __del__(self):
        self.close()

    def reset(self):
        """Reset the parser, so it may be used on a fresh stream."""
        self._gen = self.parse()
        self._awaiting = next(self._gen)

    def close(self):
        """Close the parser."""
        if self._gen is not None:
            self._gen.close()
            self._gen = None

    def feed(self, data):
        """
        Called with data (bytes), will yield 0 or more objects parsed
        from the stream.

        """
        pos = 0
        while pos < len(data):
            # Awaiting a read of a fixed number of bytes
            if isinstance(self._awaiting, self.read):
                # This many bytes left to read
                remaining = self._awaiting.remaining
                # Bite off remaining bytes
                chunk = data[pos:pos + remaining]
                chunk_size = len(chunk)
                pos += chunk_size
                # Add to buffer
                self._buffer.append(chunk)
                remaining -= chunk_size
                if remaining:
                    # Await more bytes
                    self._awaiting = self.read(remaining)
                else:
                    # Got all the bytes we need in buffer
                    send_bytes = b''.join(self._buffer)
                    del self._buffer[:]
                    # Send to coroutine, get new 'awaitable'
                    self._awaiting = self._gen.send(send_bytes)
            # Awaiting a read until a terminator
            elif isinstance(self._awaiting, self.read_until):
                # Reading to separator
                chunk = data[pos:]
                self._until += chunk
                sep = self._awaiting.sep
                sep_index = self._until.find(sep)
                if sep_index == -1:
                    # Separator not found, advance position
                    pos += len(chunk)
                else:
                    # Found separator
                    # Get data prior to and including separator
                    split_index = sep_index + len(sep)
                    send_bytes = self._until[:split_index]
                    # Reset data, to continue parsing
                    data = self._until[split_index:]
                    self._until = b''
                    pos = 0
                    # Send bytes to coroutine, get new 'awaitable'
                    self._awaiting = self._gen.send(send_bytes)
            # Yield any non-awaitables...
            while not isinstance(self._awaiting, self._Awaitable):
                yield self._awaiting
                self._awaiting = next(self._gen)


    def parse(self):
        """
        A generator to parse incoming stream.

        Yield the result of `self.read` to read n bytes from the stream.
        Yield any parsed objects

        Here's an example::

            size = struct.unpack('<I', yield self.read(4))
            data = yield self.read(size)
            yield data.decode()

        """
        return
        yield


if __name__ == "__main__":
    class TestParser(Parser):
        def parse(self):
            data = yield self.read_until(b'\r\n\r\n')
            yield data
            data = yield self.read(1)
            yield data
            data = yield self.read(2)
            yield data
            data = yield self.read(4)
            yield data
            data = yield self.read(2)
            yield data
    parser = TestParser()
    for b in (b'head', b'ers: example', b'\r\n', b'\r\n', b'12', b'34', b'5', b'678', b'', b'90'):
        for frame in parser.feed(b):
            print(repr(frame))



"""
Events emitted from Lomond Websockets.

Events may be distinguished either by their type or by the `name`
attribute. For example:

    if isinstance(event, events.PING)

or

    if event.name == 'ping'


All events have a `received_time` attribute which is the epoch time
the event was created from the network stream.

"""

from __future__ import unicode_literals

import time


class Event(object):
    """Base class for a websocket 'event'."""
    __slots__ = ['received_time']

    def __init__(self):
        self.received_time = time.time()

    def __repr__(self):
        return "{}()".format(self.__class__.__name__)

    def _summarize_bytes(cls, data, max_len=16):
        """Avoid spamming logs by truncating byte strings in repr."""
        if len(data) > max_len:
            return "{!r} + {} bytes".format(
                data[:max_len],
                len(data) - max_len
            )
        return repr(data)

    def _summarize_text(cls, text, max_len=16):
        """Avoid spamming logs by truncating text."""
        if len(text) > max_len:
            return "{!r} + {} chars".format(
                text[:max_len],
                len(text) - max_len
            )
        return repr(text)


class Poll(Event):
    """A generated poll event."""
    name = 'poll'


class Connecting(Event):
    """Connection process has started."""
    __slots__ = ['url']
    name = 'connecting'

    def __init__(self, url):
        self.url = url
        super(Connecting, self).__init__()

    def __repr__(self):
        return "{}(url='{}')".format(self.__class__.__name__, self.url)


class ConnectFail(Event):
    """Connection failed (connectivity related)."""
    __slots__ = ['reason']
    name = 'connect_fail'

    def __init__(self, reason):
        self.reason = reason
        super(ConnectFail, self).__init__()

    def __repr__(self):
        return "{}('{}')".format(
            self.__class__.__name__,
            self.reason,
        )


class Connected(Connecting):
    """Connected to the server (but not yet negotiated websockets)."""
    name = 'connected'


class Rejected(Event):
    """Server rejected WS connection."""
    __slots__ = ['response', 'reason']
    name = 'rejected'

    def __init__(self, response, reason):
        self.response = response
        self.reason = reason
        super(Rejected, self).__init__()

    def __repr__(self):
        return "{}({!r}, '{}')".format(
            self.__class__.__name__,
            self.response,
            self.reason
        )


class Ready(Event):
    """Server accepted WS connection."""
    __slots__ = ['response', 'protocol', 'extensions']
    name = 'ready'

    def __init__(self, response, protocol, extensions):
        self.response = response
        self.protocol = protocol
        self.extensions = extensions
        super(Ready, self).__init__()

    def __repr__(self):
        return '{}({!r}, protocol={!r}, extensions={!r})'.format(
            self.__class__.__name__,
            self.response,
            self.protocol,
            self.extensions
        )


class Disconnected(Event):
    """Server disconnected."""
    __slots__ = ['graceful', 'reason']
    name = 'disconnected'

    def __init__(self, reason='closed', graceful=False):
        self.reason = reason
        self.graceful = graceful

    def __repr__(self):
        return "{}('{}', graceful={!r})".format(
            self.__class__.__name__,
            self.reason,
            self.graceful
        )


class Closed(Event):
    """Websocket connection is closed."""
    __slots__ = ['code', 'reason']
    name = 'closed'

    def __init__(self, code, reason):
        self.code = code
        self.reason = reason
        super(Closed, self).__init__()

    def __repr__(self):
        return '{}({!r}, {!r})'.format(
            self.__class__.__name__,
            self.code,
            self.reason,
        )


class UnknownMessage(Event):
    """
    An application message was received, with an unknown
    opcode.
    """
    __slots__ = ['message']
    name = 'unknown'

    def __init__(self, message):
        self.message = message
        super(UnknownMessage, self).__init__()


class Binary(Event):
    """An application message was received."""
    __slots__ = ['data']
    name = 'binary'

    def __init__(self, data):
        self.data = data
        super(Binary, self).__init__()

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__,
            self._summarize_bytes(self.data)
        )


class Pong(Event):
    """An application message was received."""
    __slots__ = ['data']
    name = 'pong'

    def __init__(self, data):
        self.data = data
        super(Pong, self).__init__()

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self.data)


class Text(Event):
    """An application text message was received."""
    __slots__ = ['text']
    name = 'text'

    def __init__(self, text):
        self.text = text
        super(Text, self).__init__()

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__,
            self._summarize_text(self.text)
        )


class BackOff(Event):
    """Unable to connect, so the client will wait and try again."""
    __slots__ = ['delay']
    name = 'back_off'

    def __init__(self, delay):
        self.delay = delay
        super(BackOff, self).__init__()

    def __repr__(self):
        return "{}(delay={:0.1f})".format(
            self.__class__.__name__,
            self.delay
        )



from __future__ import unicode_literals

from ._version import __version__

# A constant used in websocket handshake
# Note, this isn't anything that needs to be secured.
# See https://tools.ietf.org/html/rfc6455#page-7
WS_KEY = b'258EAFA5-E914-47DA-95CA-C5AB0DC85B11'

# Standard since 2011
# https://tools.ietf.org/html/rfc6455
WS_VERSION = 13

# Only report major.minor in version
_version_identifier = '.'.join(__version__.split('.')[:2])

# User agent sent with websocket request
# Default will be DataplicityLomond/<major>.<minor
# e.g "DataplicityLomond/0.1"
USER_AGENT = 'DataplicityLomond/{}'.format(_version_identifier)



from __future__ import unicode_literals


class Error(Exception):
    """Base exception."""
    def __init__(self, msg, *args, **kwargs):
        error_msg = msg.format(*args, **kwargs)
        super(Error, self).__init__(error_msg)


class FrameBuildError(Error):
    """Raised when trying to build an invalid websocket frame."""


class HandshakeError(Error):
    """
    Raised when the server doesn't respond correctly to the websocket
    handshake.

    """


class ProtocolError(Error):
    """Raised in response to a protocol violation."""
    # Results in a a graceful disconnect.


class CriticalProtocolError(Error):
    """Critical protocol error."""
    # An egregious error in the protocol resulting in an immediate
    # disconnect.


class PayloadTooLarge(ProtocolError):
    """The payload length field is too large."""
    # With a max payload of 2**63 bytes, we would run out of memory
    # way before we have to raise this.


class TransportFail(Error):
    """The transport (socket) failed when sending."""
    # Likely indicates the socket failed


class WebSocketUnavailable(Error):
    """The websocket may not be used."""


class WebSocketClosed(WebSocketUnavailable):
    """Raised when attempting to send over a closed websocket."""


class WebSocketClosing(WebSocketUnavailable):
    """Raised when attempting to send over a closing websocket."""
